#include"object.hpp"
#include"wirerenderer.hpp"
#include"sincos.hpp"




namespace{


constexpr Point3D
null_point;


void
draw_point(Drawer&  dr, pixel32_t  pixel, int  x, int  y)
{
    if((x >=               0) &&
       (x <   dr.get_width()) &&
       (y >=               0) &&
       (y <  dr.get_height()))
    {
      dr.get_pixel(x,y) = pixel;
    }
}


void
draw_point(Drawer&  dr, pixel32_t  pixel, int  x, int  y, int  z)
{
  Point2D  pt(x,y,z);

  pt.x += WireRenderer::rendparam.x_offset;
  pt.y += WireRenderer::rendparam.y_offset;

  draw_point(dr,pixel,pt.x,pt.y);
}


template<typename  T>
T
zsgn(T  t)
{
  return((t < 0)? -1:
         (t > 0)?  1:0);
}


void
draw_line(Drawer&  dr, pixel32_t  pixel, int x0, int y0, int x1, int y1)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
   int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
   int err = dx+dy, e2; /* error value e_xy */
 
   for(;;){  /* loop */
      draw_point(dr,pixel,x0,y0);
      if (x0==x1 && y0==y1) break;
      e2 = 2*err;
      if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
      if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
   }
}


}





void
Node::
draw_line(Drawer&  dr, pixel32_t  pixel, int  x1, int  y1, int  z1, int  x2, int  y2, int  z2)
{
  Point2D  a(x1,y1,z1);
  Point2D  b(x2,y2,z2);

  const int  x_base = WireRenderer::rendparam.x_offset;
  const int  y_base = WireRenderer::rendparam.y_offset;

  ::draw_line(dr,pixel,x_base+a.x,
                       y_base+a.y,
                       x_base+b.x,
                       y_base+b.y);
}


void
Node::
draw_circle(Drawer&  dr, pixel32_t  pixel, int xm, int ym, int r)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int x = -r, y = 0, err = 2-2*r; /* II. Quadrant */ 
   do {
      draw_point(dr,pixel,xm-x, ym+y); /*   I. Quadrant */
      draw_point(dr,pixel,xm-y, ym-x); /*  II. Quadrant */
      draw_point(dr,pixel,xm+x, ym-y); /* III. Quadrant */
      draw_point(dr,pixel,xm+y, ym+x); /*  IV. Quadrant */
      r = err;
      if (r <= y) err += ++y*2+1;           /* e_xy+e_y < 0 */
      if (r > x || err > y) err += ++x*2+1; /* e_xy+e_x > 0 or no 2nd y-step */
   } while (x < 0);
}


void
Node::
draw_ellipse_rect(Drawer&  dr, pixel32_t  pixel, int x0, int y0, int x1, int y1)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int a = abs(x1-x0), b = abs(y1-y0), b1 = b&1; /* values of diameter */
   long dx = 4*(1-a)*b*b, dy = 4*(b1+1)*a*a; /* error increment */
   long err = dx+dy+b1*a*a, e2; /* error of 1.step */

   if (x0 > x1) { x0 = x1; x1 += a; } /* if called with swapped points */
   if (y0 > y1) y0 = y1; /* .. exchange them */
   y0 += (b+1)/2; y1 = y0-b1;   /* starting pixel */
   a *= 8*a; b1 = 8*b*b;

   do {
       draw_point(dr,pixel,x1, y0); /*   I. Quadrant */
       draw_point(dr,pixel,x0, y0); /*  II. Quadrant */
       draw_point(dr,pixel,x0, y1); /* III. Quadrant */
       draw_point(dr,pixel,x1, y1); /*  IV. Quadrant */
       e2 = 2*err;
       if (e2 <= dy) { y0++; y1--; err += dy += a; }  /* y step */ 
       if (e2 >= dx || 2*err > dy) { x0++; x1--; err += dx += b1; } /* x step */
   } while (x0 <= x1);
   
   while (y0-y1 < b) {  /* too early stop of flat ellipses a=1 */
       draw_point(dr,pixel,x0-1, y0); /* -> finish tip of ellipse */
       draw_point(dr,pixel,x1+1, y0++); 
       draw_point(dr,pixel,x0-1, y1);
       draw_point(dr,pixel,x1+1, y1--); 
   }
}


Node::
Node():
parent(nullptr),
pixel(0xFFFFFFFF),
terminate_flag(0)
{
  std::memset(&h_angle,0,sizeof(h_angle));
  std::memset(&v_angle,0,sizeof(v_angle));
}


Node::
~Node()
{
    for(auto  nd: children)
    {
      delete nd;
    }
}




void
Node::
append(Node&  child)
{
  child.parent = this;

  children.emplace_back(&child);
}


void
Node::
cleanup_as_root(const Point3D&  center_point)
{
  Point3D  tmp_point;

  rotate_horizontally(offset_point,center_point,0,       tmp_point);
    rotate_vertically(   tmp_point,center_point,0,calculated_point);

    for(auto  p: children)
    {
      p->cleanup_as_child(*this);
    }
}


void
Node::
cleanup_as_child(const Node&  root)
{
  const Point3D&  center_point = root.calculated_point;

  Point3D  tmp_point;

  rotate_horizontally(offset_point,center_point,0,       tmp_point);
    rotate_vertically(   tmp_point,center_point,0,calculated_point);

    for(auto  p: children)
    {
      p->cleanup_as_child(root);
    }
}


void
Node::
draw_as_root(Drawer&  dr, const Point3D&  center_point, int  h_angle_base, int  v_angle_base)
{
  h_angle.calculated_value = (h_angle_base+h_angle.relative_value);
  v_angle.calculated_value = (v_angle_base+v_angle.relative_value);

  Point3D  tmp_point;

  rotate_horizontally(offset_point,center_point,h_angle.calculated_value,       tmp_point);
    rotate_vertically(   tmp_point,center_point,v_angle.calculated_value,calculated_point);


  Point2D  pt(calculated_point);

  pt.x += WireRenderer::rendparam.x_offset;
  pt.y += WireRenderer::rendparam.y_offset;

  draw_circle(dr,pixel,pt.x,pt.y,circle_radius  );
  draw_circle(dr,pixel,pt.x,pt.y,circle_radius/2);

    for(auto  p: children)
    {
      p->draw_as_child(dr,*this);
    }
}


void
Node::
draw_as_child(Drawer&  dr, const Node&  root)
{
//  h_angle.calculated_value = (parent->h_angle.calculated_value+h_angle.relative_value);
//  v_angle.calculated_value = (parent->v_angle.calculated_value+v_angle.relative_value);
  h_angle.calculated_value = (parent->h_angle.calculated_value);
  v_angle.calculated_value = (parent->v_angle.calculated_value);


  auto&  start_point = parent->calculated_point;

  Point3D  tmp_point;

  rotate_horizontally(offset_point,null_point,h_angle.calculated_value,       tmp_point);
    rotate_vertically(   tmp_point,null_point,v_angle.calculated_value,calculated_point);

  calculated_point += start_point;


  Point2D  pt(calculated_point);

  pt.x += WireRenderer::rendparam.x_offset;
  pt.y += WireRenderer::rendparam.y_offset;

  draw_circle(dr,pixel,pt.x,pt.y,circle_radius);



  draw_line(dr,pixel,start_point.x,
                     start_point.y,
                     start_point.z,
                     calculated_point.x,
                     calculated_point.y,
                     calculated_point.z);


    for(auto  p: children)
    {
      p->draw_as_child(dr,root);
    }
}


void
Node::
draw_as_stepchild(Drawer&  dr, const Node&  root, const Node&  stepparent)
{
  h_angle.calculated_value = (stepparent.h_angle.calculated_value+h_angle.relative_value);
  v_angle.calculated_value = (stepparent.v_angle.calculated_value+v_angle.relative_value);


  auto&  start_point = stepparent.calculated_point;

  Point3D  calculated_center_point = root.calculated_point;

  Point3D  tmp_point;

  rotate_horizontally(offset_point,null_point,h_angle.calculated_value,       tmp_point);
    rotate_vertically(   tmp_point,null_point,v_angle.calculated_value,calculated_point);

  calculated_point += start_point;


  Point2D  pt(calculated_point);

  pt.x += WireRenderer::rendparam.x_offset;
  pt.y += WireRenderer::rendparam.y_offset;

  draw_circle(dr,pixel,pt.x,pt.y,circle_radius);


  draw_line(dr,pixel,start_point.x,
                     start_point.y,
                     start_point.z,
                     calculated_point.x,
                     calculated_point.y,
                     calculated_point.z);


    for(auto  p: children)
    {
      p->draw_as_child(dr,root);
    }
}




